/**************************************************************************/
/*                                                                        */
/* Copyright (c) 2001, 2006 NoMachine, http://www.nomachine.com.          */
/*                                                                        */
/* NXCLIENT, NX protocol compression and NX extensions to this software   */
/* are copyright of NoMachine. Redistribution and use of the present      */
/* software is allowed according to terms specified in the file LICENSE   */
/* which comes in the source distribution.                                */
/*                                                                        */
/* Check http://www.nomachine.com/licensing.html for applicability.       */
/*                                                                        */
/* NX and NoMachine are trademarks of Medialogic S.p.A.                   */
/*                                                                        */
/* All rights reserved.                                                   */
/*                                                                        */
/**************************************************************************/


#include <System.h>
#include <Logger.h>
#include <Display.h>
#include "ServiceDisplay.h"

#include <sys/un.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <time.h>
#include <sys/stat.h>
#include <sys/types.h>

#ifdef __CYGWIN__
extern "C" void cygwin_conv_to_full_posix_path(const char *path, char *posix_path);
extern "C" void cygwin_win32_to_posix_path_list(const char *win32, char *posix);
#endif

ServiceDisplay::ServiceDisplay(const char* path, int argc, char** argv)
  :Service(path, "display")
{
   resolution_ = "640x480";
   name_ = "NX";
   multiwindow_ = 0;
   fullscreen_ = 0;
   fontserver_ = NULL;

   serverpath_ = NULL;
   authpath_ = NULL;
   display_ = NULL;

   int n = 0;

   #ifdef __CYGWIN__
   for (n = 0; n < argc; n++)
   {
     if (strcmp(argv[n], "--name") == 0)
     {
       if (n+1 < argc) name_ = strdup(argv[n+1]);
     }
     
     if (strcmp(argv[n], "--fontserver") == 0)
     {
       if (n+1 < argc) fontserver_ = strdup(argv[n+1]);
     }
    
 
     else if (strcmp(argv[n], "--resolution") == 0)
     {
       if (n+1 < argc)
       {
	 if (strcmp(argv[n+1], "fullscreen") == 0)
	 {
	    fullscreen_ = 1;
	 }
	 else
	   resolution_ = strdup(argv[n+1]);
       }
     }
     else if (strcmp(argv[n], "--multiwindow") == 0)
     {
       if (n+1 < argc && strcmp(argv[n+1], "1") == 0)
       {
         multiwindow_ = 1;
       }
     }
   }

   #else

   for (n = 0; n < argc; n++)
   {
     if (strcmp(argv[n], "--serverpath") == 0)
     {
       if (n+1 < argc) serverpath_ = strdup(argv[n+1]);
     }
     else if (strcmp(argv[n], "--auth") == 0)
     {
       if (n+1 < argc) authpath_ = strdup(argv[n+1]);
     }
     else if (strcmp(argv[n], "--display") == 0)
     {
       if (n+1 < argc) display_ = strdup(argv[n+1]);
     }
   }

   #endif
}

ServiceDisplay::~ServiceDisplay()
{
}

int ServiceDisplay::start()
{
  const char *systemPath = getSystemDir();
  char s_display[DEFAULT_STRING_LENGTH];

  #ifdef __CYGWIN__

  int display = 0;

  //
  // FIXME: we do not anything in case of failure
  //
  writeTypeFile();

  if (systemPath == NULL)
  {
    logUser("Error: Cannot find NX system path. Exiting.");
    return -1;
  }

  if ((strlen(systemPath) + strlen("/bin/") + strlen(DISPLAY_BIN) + 1) > DEFAULT_STRING_LENGTH)
  {
    logUser("Error: nxesd path too long. Exiting.\n");
    return -1;
  }

  strcpy(executablePath_, systemPath);
  strcat(executablePath_, "/bin/");
  strcat(executablePath_, DISPLAY_BIN);

  port_ = getFreePort(6000);

  if (port_ == 0)
  {
    logUser("Error: Cannot find free port. Exiting.\n");
    return -1;
  }

  display = getFreeDisplay();

  logUser("Info: Using display: %d", display);

  char *authorityPath = getenv("XAUTHORITY");

  logUser("Info: Using authority file: %s", authorityPath);

  #define FONT_PATHS_MAX 10000

  char *fontPaths = new char[FONT_PATHS_MAX];
  if(fontserver_)
  {
     strcpy(fontPaths, fontserver_);
  }
  else
  {
     const char *paths[] = { "\\X11R6\\lib\\X11\\fonts\\TTF",
        "\\X11R6\\lib\\X11\\fonts\\base",
        "\\X11R6\\lib\\X11\\fonts\\misc",
        "\\X11R6\\lib\\X11\\fonts\\Speedo",
        "\\X11R6\\lib\\X11\\fonts\\Type1",
        "\\X11R6\\lib\\X11\\fonts\\75dpi",
        "\\X11R6\\lib\\X11\\fonts\\100dpi",
        "\\usr\\X11R6\\lib\\X11\\fonts\\TTF",
        "\\usr\\X11R6\\lib\\X11\\fonts\\base",
        "\\usr\\X11R6\\lib\\X11\\fonts\\misc",
        "\\usr\\X11R6\\lib\\X11\\fonts\\Speedo",
        "\\usr\\X11R6\\lib\\X11\\fonts\\Type1",
        "\\usr\\X11R6\\lib\\X11\\fonts\\75dpi",
        "\\usr\\X11R6\\lib\\X11\\fonts\\100dpi"};

     //int len = FONT_PATHS_MAX;

     strcpy(fontPaths, "\"");

     for (int i = 0;i < 14; i++)
     {
       if (i) strcat(fontPaths, ",");

       char path[255];
       char posixPath[255];

       strcpy(path, getSystemDir());
       strcat(path, "\\");
       strcat(path, paths[i]);
       cygwin_win32_to_posix_path_list(path, posixPath);
       logUser("Posix path=%s path=%s", posixPath, path);
       strcat(fontPaths, posixPath);
     }

     strcat(fontPaths, "\"");
   }
   
   logUser("fontPaths=%s", fontPaths);
 
   snprintf(s_display, DEFAULT_STRING_LENGTH, ":%d", display);


   addArgument(executablePath_);
   addArgument("-nowinkill");
   addArgument("-clipboard");
   addArgument("-noloadxkb");
   addArgument("-agent");
   addArgument("-hide");
   addArgument("-noreset");
   addArgument("-auth");
   addArgument(authorityPath);
   addArgument("-nolisten");
   addArgument("tcp");

   addArgument("-fp");
   addArgument(fontPaths);

   if (fullscreen_)
   {
     addArgument("-fullscreen");
   }
   else
   {
     addArgument("-screen");
     addArgument("0");
     addArgument(resolution_);
   }

   if (multiwindow_)
   {
     addArgument("-multiwindow");
   }

   addArgument("-name");
   addArgument(name_);
   addArgument(s_display);

   #else //__CYGWIN__

   if (!serverpath_)
   {
     logUser("Error: Path to X Apple server is empty.");
     return -1;
   }

   if (!authpath_)
   {
     logUser("Error: Path to authority file is empty.");
     return -1;
   }

   if (!display_)
   {
     logUser("Error: Display value is empty.");
     return -1;
   }

   strcpy(executablePath_, serverpath_);

   logUser("Info: Starting AppleXServer[%s] on display [%s]", executablePath_, display_);

   addArgument(executablePath_);
   addArgument("-auth");
   addArgument(authpath_);
   addArgument(display_);

   #endif //

   run();

   return 0;
}

int ServiceDisplay::getFreeDisplay()
{
  sockaddr_un xServerAddrUNIX;

  int xServerAddrFamily = AF_UNIX;

  xServerAddrUNIX.sun_family = AF_UNIX;

  for (int d = 0; d < 100; d++)
  {
    //logUser("getFreeDisplay: %d", d);
     //for mac we have to use this:
     //QString displayPath = "/tmp/.X11-unix/X" + QString::number(d);
     //for windows:
     char displayPath[DEFAULT_STRING_LENGTH];
     char displayStr[DEFAULT_STRING_LENGTH];

     #ifdef __CYGWIN__
     strcpy(displayPath, rootPath_);
     #else
     strcpy(displayPath, "/tmp");
     #endif
     strcat(displayPath, "/.X11-unix");

     /*
      * let's try to create the directory
      */

     mkdir(displayPath, S_IRUSR|S_IWUSR|S_IXUSR);

     /*
      * let's check if we can find free display - this will be probably alway 1st one
      */

     strcat(displayPath, "/X");

     snprintf(displayStr, DEFAULT_STRING_LENGTH, "%d", d);
     strncat(displayPath, displayStr, DEFAULT_STRING_LENGTH);

     logUser("Info: getFreeDisplay: path=%s", displayPath);

     strcpy(xServerAddrUNIX.sun_path, displayPath);

     sockaddr *xServerAddr = (sockaddr *) & xServerAddrUNIX;
     unsigned int xServerAddrLength = sizeof(sockaddr_un);

     //
     // Connect to the X server.
     //

     int xServerFd;

     xServerFd = socket(xServerAddrFamily, SOCK_STREAM, PF_UNSPEC);

     if (xServerFd < 0)
     {
	/*
	 * It is serious problem we should quit
	 */

	logUser("PANIC! Failed to create the socket. Error is [%d], [%s].", errno, strerror(errno));

	return -1;
     }

     int result = connect(xServerFd, xServerAddr, xServerAddrLength);

     if (result < 0)
     {
	/*
	 * it means that there is no server running on that socket we can start new
	 * server
	 */

	logUser("Failed to connect to the X server on [:%i]. Error is [%d], [%s].", d, errno, strerror(errno));
	close(xServerFd);
	return d;
     }
     else
     {
	/*
	 * it means that the server is running so we should skip and use
	 * another socket
	 */
	logUser("Successfully connected to the X server on [:%i].", d);
     }

     close(xServerFd);
  }

  return -1;
}




